@@ -1,8 +1,8 @@
 /*
 **  OSSP fsl - Faking/Flexible Syslog Library
-**  Copyright (c) 2002-2004 Ralf S. Engelschall <rse@engelschall.com>
-**  Copyright (c) 2002-2004 The OSSP Project <http://www.ossp.org/>
-**  Copyright (c) 2002-2004 Cable & Wireless Deutschland <http://www.cw.com/de/>
+**  Copyright (c) 2002-2005 Ralf S. Engelschall <rse@engelschall.com>
+**  Copyright (c) 2002-2005 The OSSP Project <http://www.ossp.org/>
+**  Copyright (c) 2002-2005 Cable & Wireless <http://www.cw.com/>
 **
 **  This file is part of OSSP fsl, a syslog(3) API faking library which
 **  can be found at http://www.ossp.org/pkg/lib/fsl/.
@@ -43,6 +43,9 @@
 #include <dirent.h>
 
 /* standard include we re-implement */
+#ifdef __hpux
+#define _XOPEN_SOURCE_EXTENDED
+#endif
 #include <syslog.h>
 
 /* third party (linked in) */
