@@ -1,8 +1,8 @@
 /*
 **  OSSP cfg - Configuration Parsing
-**  Copyright (c) 1999-2002 Ralf S. Engelschall <rse@engelschall.com>
-**  Copyright (c) 1999-2002 The OSSP Project (http://www.ossp.org/)
-**  Copyright (c) 2001-2002 Cable & Wireless Deutschland (http://www.cw.com/de/)
+**  Copyright (c) 2002-2005 Ralf S. Engelschall <rse@engelschall.com>
+**  Copyright (c) 2002-2005 The OSSP Project (http://www.ossp.org/)
+**  Copyright (c) 2002-2005 Cable & Wireless (http://www.cw.com/)
 **
 **  This file is part of OSSP cfg, a configuration parsing
 **  library which can be found at http://www.ossp.org/pkg/lib/cfg/.
@@ -35,13 +35,17 @@
 #include "cfg.h"
 #include "cfg_buf.h"
 #include "cfg_fmt.h"
+#include "cfg_global.h"
 
+/* opaque buffer data structure */
 struct cfg_buf_st {
-    char   *buf_ptr;
-    size_t  buf_size;
-    size_t  buf_len;
+    char   *buf_ptr;      /* pointer to buffer start              */
+    size_t  buf_size;     /* total size of buffer in bytes        */
+    size_t  buf_len;      /* character length of string in buffer */
+                          /* (not counting NUL-termination)       */
 };
 
+/* create a new buffer object */
 cfg_rc_t cfg_buf_create(cfg_buf_t **buf)
 {
     if (buf == NULL)
@@ -54,6 +58,7 @@ cfg_rc_t cfg_buf_create(cfg_buf_t **buf)
     return CFG_OK;
 }
 
+/* resize (destroy, expand or shrink) the buffer */
 cfg_rc_t cfg_buf_resize(cfg_buf_t *buf, signed int n)
 {
     char *cp;
@@ -61,7 +66,7 @@ cfg_rc_t cfg_buf_resize(cfg_buf_t *buf, signed int n)
     if (buf == NULL)
         return CFG_ERR_ARG;
     if (n == 0) {
-        /* destroy buffer */
+        /* special case: destroy buffer */
         if (buf->buf_ptr != NULL)
             free(buf->buf_ptr);
         buf->buf_ptr  = NULL;
@@ -75,6 +80,7 @@ cfg_rc_t cfg_buf_resize(cfg_buf_t *buf, signed int n)
                 return CFG_ERR_SYS;
             buf->buf_size = n+1;
             buf->buf_len  = 0;
+            *buf->buf_ptr = '\0';
         }
         else {
             if ((cp = realloc(buf->buf_ptr, buf->buf_size+n)) == NULL)
@@ -90,6 +96,46 @@ cfg_rc_t cfg_buf_resize(cfg_buf_t *buf, signed int n)
     return CFG_OK;
 }
 
+/* append a string and/or single character to the buffer */
+cfg_rc_t cfg_buf_append(cfg_buf_t *buf, const char *str, size_t len, char c)
+{
+    cfg_rc_t rc;
+
+    if (buf == NULL)
+        return CFG_ERR_ARG;
+    if (str != NULL) {
+        if (len == 0)
+            len = strlen(str);
+        if ((rc = cfg_buf_resize(buf, len)) != CFG_OK)
+            return rc;
+        memcpy(buf->buf_ptr + buf->buf_len, str, len);
+        buf->buf_len += len;
+    }
+    if (c != '\0') {
+        if ((rc = cfg_buf_resize(buf, 1)) != CFG_OK)
+            return rc;
+        *(buf->buf_ptr + buf->buf_len) = c;
+        buf->buf_len++;
+    }
+    *(buf->buf_ptr + buf->buf_len) = '\0';
+    return CFG_OK;
+}
+
+/* remove a trailing string from the buffer */
+cfg_rc_t cfg_buf_remove(cfg_buf_t *buf, const char *str, size_t len)
+{
+    if (buf == NULL || len == 0)
+        return CFG_ERR_ARG;
+    if (len > buf->buf_len)
+        return CFG_ERR_USE;
+    if (str != NULL)
+        memcpy((void *)str, buf->buf_ptr + buf->buf_len - len, len + 1);
+    buf->buf_len -= len;
+    *(buf->buf_ptr + buf->buf_len) = '\0';
+    return CFG_OK;
+}
+
+/* append a formatted string to the buffer */
 cfg_rc_t cfg_buf_format(cfg_buf_t *buf, const char *fmt, ...)
 {
     va_list ap;
@@ -101,6 +147,7 @@ cfg_rc_t cfg_buf_format(cfg_buf_t *buf, const char *fmt, ...)
     return rc;
 }
 
+/* append a formatted string to the buffer (va_list based) */
 cfg_rc_t cfg_buf_vformat(cfg_buf_t *buf, const char *fmt, va_list ap)
 {
     cfg_rc_t rc;
@@ -112,27 +159,37 @@ cfg_rc_t cfg_buf_vformat(cfg_buf_t *buf, const char *fmt, va_list ap)
         return CFG_ERR_FMT;
     if ((rc = cfg_buf_resize(buf, n)) != CFG_OK)
         return rc;
-    if ((n = cfg_fmt_vsprintf(buf->buf_ptr + buf->buf_len, 
-                              buf->buf_size - buf->buf_len, 
+    if ((n = cfg_fmt_vsprintf(buf->buf_ptr  + buf->buf_len,
+                              buf->buf_size - buf->buf_len,
                               fmt, ap)) == -1)
         return CFG_ERR_FMT;
     buf->buf_len += n;
     return CFG_OK;
 }
 
+/* return the buffer string */
 cfg_rc_t cfg_buf_content(cfg_buf_t *buf, char **ptr, size_t *len, size_t *size)
 {
     if (buf == NULL)
         return CFG_ERR_ARG;
-    if (ptr != NULL)
-        *ptr  = buf->buf_ptr;
     if (len != NULL)
         *len  = buf->buf_len;
     if (size != NULL)
         *size = buf->buf_size;
+    if (ptr != NULL) {
+        if (buf->buf_ptr == NULL)
+            *ptr = strdup("");
+        else {
+            *ptr = buf->buf_ptr;
+            buf->buf_ptr  = NULL;
+            buf->buf_size = 0;
+            buf->buf_len  = 0;
+        }
+    }
     return CFG_OK;
 }
 
+/* destroy buffer object */
 cfg_rc_t cfg_buf_destroy(cfg_buf_t *buf)
 {
     if (buf == NULL)
