@@ -1,8 +1,8 @@
 /*
 **  OSSP cfg - Configuration Parsing
-**  Copyright (c) 2002-2003 Ralf S. Engelschall <rse@engelschall.com>
-**  Copyright (c) 2002-2003 The OSSP Project (http://www.ossp.org/)
-**  Copyright (c) 2002-2003 Cable & Wireless Deutschland (http://www.cw.com/de/)
+**  Copyright (c) 2002-2005 Ralf S. Engelschall <rse@engelschall.com>
+**  Copyright (c) 2002-2005 The OSSP Project (http://www.ossp.org/)
+**  Copyright (c) 2002-2005 Cable & Wireless (http://www.cw.com/)
 **
 **  This file is part of OSSP cfg, a configuration parsing
 **  library which can be found at http://www.ossp.org/pkg/lib/cfg/.
@@ -38,6 +38,8 @@ typedef struct cfg_buf_st cfg_buf_t;
 
 extern cfg_rc_t cfg_buf_create  (cfg_buf_t **buf);
 extern cfg_rc_t cfg_buf_resize  (cfg_buf_t  *buf, signed int n);
+extern cfg_rc_t cfg_buf_append  (cfg_buf_t  *buf, const char *str, size_t len, char c);
+extern cfg_rc_t cfg_buf_remove  (cfg_buf_t  *buf, const char *str, size_t len);
 extern cfg_rc_t cfg_buf_format  (cfg_buf_t  *buf, const char *fmt, ...);
 extern cfg_rc_t cfg_buf_vformat (cfg_buf_t  *buf, const char *fmt, va_list ap);
 extern cfg_rc_t cfg_buf_content (cfg_buf_t  *buf, char **ptr, size_t *len, size_t *size);
