@@ -1,8 +1,8 @@
 /*
 **  OSSP cfg - Configuration Parsing
-**  Copyright (c) 1999-2002 Ralf S. Engelschall <rse@engelschall.com>
-**  Copyright (c) 1999-2002 The OSSP Project (http://www.ossp.org/)
-**  Copyright (c) 2001-2002 Cable & Wireless Deutschland (http://www.cw.com/de/)
+**  Copyright (c) 2002-2005 Ralf S. Engelschall <rse@engelschall.com>
+**  Copyright (c) 2002-2005 The OSSP Project (http://www.ossp.org/)
+**  Copyright (c) 2002-2005 Cable & Wireless (http://www.cw.com/)
 **
 **  This file is part of OSSP cfg, a configuration parsing
 **  library which can be found at http://www.ossp.org/pkg/lib/cfg/.
@@ -39,6 +39,7 @@
 #include "cfg_data.h"
 #include "cfg_node.h"
 #include "cfg_syn.h"
+#include "cfg_global.h"
 #define _CFG_VERS_C_AS_HEADER_
 #include "cfg_vers.c"
 #undef  _CFG_VERS_C_AS_HEADER_
@@ -72,10 +73,18 @@ cfg_rc_t cfg_create(cfg_t **cfg)
     return CFG_OK;
 }
 
+static cfg_rc_t cfg_destroy_node(cfg_t *cfg, cfg_node_t *node, void *ctx)
+{
+    cfg_node_destroy(cfg, node);
+    return CFG_OK;
+}
+
 cfg_rc_t cfg_destroy(cfg_t *cfg)
 {
     if (cfg == NULL)
         return CFG_ERR_ARG;
+    if (cfg->root != NULL)
+        cfg_node_apply(cfg, cfg->root, NULL, NULL, cfg_destroy_node, NULL);
     cfg_grid_destroy(cfg->grid_nodes);
     cfg_grid_destroy(cfg->grid_tokens);
     free(cfg);
@@ -155,7 +164,7 @@ cfg_rc_t cfg_import(cfg_t *cfg, cfg_node_t *node, cfg_fmt_t fmt, const char *in_
     char error[1024];
     cfg_node_t *root;
 
-    if (cfg == NULL || in_ptr == NULL || in_len == 0)
+    if (cfg == NULL || in_ptr == NULL)
         return CFG_ERR_ARG;
     if (node == NULL)
         node = cfg->root;
@@ -166,8 +175,8 @@ cfg_rc_t cfg_import(cfg_t *cfg, cfg_node_t *node, cfg_fmt_t fmt, const char *in_
         }
     }
     /* FIXME: CFG_FMT_XML */
-    else 
-        return CFG_ERR_INT; 
+    else
+        return CFG_ERR_INT;
 
     /* FIXME: merge: cfg->root[node] := cfg->root[node] + root */
     cfg->root = root;
@@ -190,11 +199,11 @@ cfg_rc_t cfg_export(cfg_t *cfg, cfg_node_t *node, cfg_fmt_t fmt, char **ex_ptr,
             return rv;
     }
     /* FIXME: CFG_FMT_XML */
-    else 
-        return CFG_ERR_INT; 
+    else
+        return CFG_ERR_INT;
 
     /* provide output */
-    if (ex_len == 0) 
+    if (ex_len == 0)
         *ex_ptr = output;
     else {
         n = strlen(output);
