@@ -1,8 +1,8 @@
 /*
 **  OSSP cfg - Configuration Parsing
-**  Copyright (c) 1999-2002 Ralf S. Engelschall <rse@engelschall.com>
-**  Copyright (c) 1999-2002 The OSSP Project (http://www.ossp.org/)
-**  Copyright (c) 2001-2002 Cable & Wireless Deutschland (http://www.cw.com/de/)
+**  Copyright (c) 2002-2005 Ralf S. Engelschall <rse@engelschall.com>
+**  Copyright (c) 2002-2005 The OSSP Project (http://www.ossp.org/)
+**  Copyright (c) 2002-2005 Cable & Wireless (http://www.cw.com/)
 **
 **  This file is part of OSSP cfg, a configuration parsing
 **  library which can be found at http://www.ossp.org/pkg/lib/cfg/.
@@ -37,77 +37,125 @@
 #include "cfg_main.h"
 #include "cfg_node.h"
 #include "cfg_fmt.h"
+#include "cfg_global.h"
 
-cfg_rc_t cfg_node_create(cfg_t *cfg, cfg_node_t **node)
+/* create a configuration node */
+cfg_rc_t
+cfg_node_create(
+    cfg_t *cfg,
+    cfg_node_t **node)
 {
     cfg_node_t *n;
     cfg_rc_t rc;
 
+    /* argument sanity checking */
     if (node == NULL)
         return CFG_ERR_ARG;
-    if ((rc = cfg_grid_alloc(cfg->grid_nodes, (void **)&n)) != CFG_OK)
+
+    /* allocate memory */
+    if ((rc = cfg_grid_alloc(cfg->grid_nodes, (void **)(void *)&n)) != CFG_OK)
         return rc;
+
+    /* initialize node attributes */
+    n->owner    = 0;
     n->parent   = NULL;
-    n->lbroth   = NULL;
     n->rbroth   = NULL;
     n->child1   = NULL;
-    n->srcname  = NULL;
-    n->srcpos   = 0;
     n->type     = CFG_NODE_TYPE_ARG;
     n->token    = NULL;
     cfg_data_init(&n->data);
+    n->srcname  = NULL;
+    n->srcpos   = 0;
     *node = n;
+
     return CFG_OK;
 }
 
-cfg_rc_t cfg_node_destroy(cfg_t *cfg, cfg_node_t *node)
+/* destroy a configuration node */
+cfg_rc_t
+cfg_node_destroy(
+    cfg_t *cfg,
+    cfg_node_t *node)
 {
+    /* argument sanity checking */
     if (node == NULL)
         return CFG_ERR_ARG;
-    if (node->token != NULL)
+
+    /* destroy memory */
+    if ((node->owner & CFG_NODE_ATTR_TOKEN) && node->token != NULL)
         free(node->token);
-    if (node->srcname != NULL)
+    if ((node->owner & CFG_NODE_ATTR_SRCNAME) && node->srcname != NULL)
         free(node->srcname);
+    cfg_data_kill(&node->data);
     cfg_grid_free(cfg->grid_nodes, node);
+
     return CFG_OK;
 }
 
-cfg_rc_t cfg_node_clone(cfg_t *cfg, cfg_node_t *node, cfg_node_t **node2)
+/* clone a node */
+cfg_rc_t
+cfg_node_clone(
+    cfg_t *cfg,
+    cfg_node_t *node,
+    cfg_node_t **node2)
 {
     cfg_node_t *n;
     cfg_rc_t rc;
 
+    /* argument sanity checking */
     if (node == NULL || node2 == NULL)
         return CFG_ERR_ARG;
-    if ((rc = cfg_grid_alloc(cfg->grid_nodes, (void **)&n)) != CFG_OK)
+
+    /* allocate memory for new node */
+    if ((rc = cfg_grid_alloc(cfg->grid_nodes, (void **)(void *)&n)) != CFG_OK)
         return rc;
+
+    /* clone node attributes */
+    n->owner    = node->owner;
     n->parent   = node->parent;
-    n->lbroth   = node->lbroth;
     n->rbroth   = node->rbroth;
     n->child1   = node->child1;
-    n->srcname  = (node->srcname != NULL ? strdup(node->srcname) : NULL);
-    n->srcpos   = node->srcpos;
     n->type     = node->type;
-    n->token    = (node->token != NULL ? strdup(node->token) : NULL);
+    n->srcpos   = node->srcpos;
+    if (node->owner & CFG_NODE_ATTR_TOKEN)
+        n->token = (node->token != NULL ? strdup(node->token) : NULL);
+    else
+        n->token = node->token;
+    if (node->owner & CFG_NODE_ATTR_SRCNAME)
+        n->srcname = (node->srcname != NULL ? strdup(node->srcname) : NULL);
+    else
+        n->srcname = node->srcname;
     cfg_data_copy(&node->data, &n->data);
+
+    /* store result */
     *node2 = n;
+
     return CFG_OK;
 }
 
-cfg_rc_t cfg_node_set(cfg_t *cfg, cfg_node_t *node, cfg_node_attr_t attr, ...)
+/* set a node attribute */
+cfg_rc_t
+cfg_node_set(
+    cfg_t *cfg,
+    cfg_node_t *node,
+    cfg_node_attr_t attr,
+    ...)
 {
     va_list ap;
 
+    /* argument sanity checking */
     if (node == NULL)
         return CFG_ERR_ARG;
+
+    /* dispatch into individual attribute handling */
     va_start(ap, attr);
-    switch (attr) {
+    switch (attr & ~(CFG_ATTR_LOAN|CFG_ATTR_GIFT|CFG_ATTR_COPY)) {
         case CFG_NODE_ATTR_PARENT: {
             node->parent = (cfg_node_t *)va_arg(ap, cfg_node_t *);
             break;
         }
         case CFG_NODE_ATTR_LBROTH: {
-            return CFG_ERR_USE;
+            return CFG_ERR_USE; /* cannot be set, only get */
         }
         case CFG_NODE_ATTR_RBROTH: {
             node->rbroth = (cfg_node_t *)va_arg(ap, cfg_node_t *);
@@ -118,52 +166,62 @@ cfg_rc_t cfg_node_set(cfg_t *cfg, cfg_node_t *node, cfg_node_attr_t attr, ...)
             break;
         }
         case CFG_NODE_ATTR_CHILDL: {
-            return CFG_ERR_USE;
+            return CFG_ERR_USE; /* cannot be set, only get */
         }
         case CFG_NODE_ATTR_CHILDS: {
-            return CFG_ERR_USE;
+            return CFG_ERR_USE; /* cannot be set, only get */
         }
         case CFG_NODE_ATTR_NODES: {
-            return CFG_ERR_USE;
+            return CFG_ERR_USE; /* cannot be set, only get */
         }
         case CFG_NODE_ATTR_DEPTH: {
-            return CFG_ERR_USE;
-        }
-        case CFG_NODE_ATTR_SRCNAME: {
-            if (node->srcname != NULL)
-                free(node->srcname);
-            node->srcname = (char *)va_arg(ap, char *);
-            if (node->srcname != NULL)
-                node->srcname = strdup(node->srcname);
-            break;
-        }
-        case CFG_NODE_ATTR_SRCPOS: {
-            node->srcpos = (int)va_arg(ap, int);
-            break;
+            return CFG_ERR_USE; /* cannot be set, only get */
         }
         case CFG_NODE_ATTR_TYPE: {
             node->type = (cfg_node_type_t)va_arg(ap, cfg_node_type_t);
             break;
         }
         case CFG_NODE_ATTR_TOKEN: {
-            if (node->token != NULL)
+            char *value = (char *)va_arg(ap, char *);
+            if (attr & CFG_ATTR_COPY)
+                value = strdup(value);
+            if (node->owner & CFG_NODE_ATTR_TOKEN && node->token != NULL)
                 free(node->token);
-            node->token = (char *)va_arg(ap, char *);
-            if (node->token != NULL)
-                node->token = strdup(node->token);
+            node->token = value;
+            if (attr & (CFG_ATTR_COPY|CFG_ATTR_GIFT))
+                node->owner |= CFG_NODE_ATTR_TOKEN;
             break;
         }
         case CFG_NODE_ATTR_DATA: {
             return CFG_ERR_USE;
         }
+        case CFG_NODE_ATTR_SRCNAME: {
+            char *value = (char *)va_arg(ap, char *);
+            if (attr & CFG_ATTR_COPY)
+                value = strdup(value);
+            if (node->owner & CFG_NODE_ATTR_SRCNAME && node->srcname != NULL)
+                free(node->srcname);
+            node->srcname = value;
+            if (attr & (CFG_ATTR_COPY|CFG_ATTR_GIFT))
+                node->owner |= CFG_NODE_ATTR_SRCNAME;
+            break;
+        }
+        case CFG_NODE_ATTR_SRCPOS: {
+            node->srcpos = (int)va_arg(ap, int);
+            break;
+        }
         default:
             return CFG_ERR_ARG;
     }
     va_end(ap);
+
     return CFG_OK;
 }
 
-static int cfg_node_get_nodes(cfg_node_t *node)
+/* INTERNAL: calculate number of nodes */
+static int
+cfg_node_get_nodes(
+    cfg_node_t *node)
 {
     int n;
 
@@ -179,14 +237,23 @@ static int cfg_node_get_nodes(cfg_node_t *node)
     return n;
 }
 
-cfg_rc_t cfg_node_get(cfg_t *cfg, cfg_node_t *node, cfg_node_attr_t attr, ...)
+/* get a node attribute */
+cfg_rc_t
+cfg_node_get(
+    cfg_t *cfg,
+    cfg_node_t *node,
+    cfg_node_attr_t attr,
+    ...)
 {
     va_list ap;
 
+    /* argument sanity checking */
     if (node == NULL)
         return CFG_ERR_ARG;
+
+    /* dispatch into individual attribute handling */
     va_start(ap, attr);
-    switch (attr) {
+    switch (attr & ~(CFG_ATTR_LOAN|CFG_ATTR_GIFT|CFG_ATTR_COPY)) {
         case CFG_NODE_ATTR_PARENT: {
             cfg_node_t **n = (cfg_node_t **)va_arg(ap, void *);
             if (n == NULL)
@@ -260,26 +327,12 @@ cfg_rc_t cfg_node_get(cfg_t *cfg, cfg_node_t *node, cfg_node_attr_t attr, ...)
             cfg_node_t *n;
             if (k == NULL)
                 return CFG_ERR_ARG;
-            *k = 0; 
+            *k = 0;
             n = node;
             while ((n = n->parent) != NULL)
                 (*k)++;
             break;
         }
-        case CFG_NODE_ATTR_SRCNAME: {
-            char **name = (char **)va_arg(ap, char **);
-            if (name == NULL)
-                return CFG_ERR_ARG;
-            *name = node->srcname;
-            break;
-        }
-        case CFG_NODE_ATTR_SRCPOS: {
-            int *pos = (int *)va_arg(ap, int *);
-            if (pos == NULL)
-                return CFG_ERR_ARG;
-            *pos = node->srcpos;
-            break;
-        }
         case CFG_NODE_ATTR_TYPE: {
             cfg_node_type_t *type = (cfg_node_type_t *)va_arg(ap, void *);
             if (type == NULL)
@@ -292,6 +345,10 @@ cfg_rc_t cfg_node_get(cfg_t *cfg, cfg_node_t *node, cfg_node_attr_t attr, ...)
             if (token == NULL)
                 return CFG_ERR_ARG;
             *token = node->token;
+            if (attr & CFG_ATTR_COPY)
+                *token = strdup(*token);
+            if (attr & CFG_ATTR_GIFT)
+                node->token = NULL;
             break;
         }
         case CFG_NODE_ATTR_DATA: {
@@ -301,33 +358,68 @@ cfg_rc_t cfg_node_get(cfg_t *cfg, cfg_node_t *node, cfg_node_attr_t attr, ...)
             *data = &(node->data);
             break;
         }
+        case CFG_NODE_ATTR_SRCNAME: {
+            char **name = (char **)va_arg(ap, char **);
+            if (name == NULL)
+                return CFG_ERR_ARG;
+            *name = node->srcname;
+            if (attr & CFG_ATTR_COPY)
+                *name = strdup(*name);
+            if (attr & CFG_ATTR_GIFT)
+                node->srcname = NULL;
+            break;
+        }
+        case CFG_NODE_ATTR_SRCPOS: {
+            int *pos = (int *)va_arg(ap, int *);
+            if (pos == NULL)
+                return CFG_ERR_ARG;
+            *pos = node->srcpos;
+            break;
+        }
         default:
             return CFG_ERR_ARG;
     }
     va_end(ap);
+
     return CFG_OK;
 }
 
-cfg_rc_t cfg_node_root(cfg_t *cfg, cfg_node_t **node)
+/* get and/or set root node */
+cfg_rc_t
+cfg_node_root(
+    cfg_t *cfg,
+    cfg_node_t  *node,
+    cfg_node_t **node_old)
 {
-    if (cfg == NULL || node == NULL)
+    /* argument sanity checking */
+    if (cfg == NULL || (node == NULL && node_old == NULL))
         return CFG_ERR_ARG;
-    *node = cfg->root;
+
+    /* optionally get old root node pointer */
+    if (node_old != NULL)
+        *node_old = cfg->root;
+
+    /* optionally set new root node pointer */
+    if (node != NULL)
+        cfg->root = node;
+
     return CFG_OK;
 }
 
-static cfg_rc_t 
+/* forward declarations for internal stepping functions */
+static cfg_rc_t
 cfg_node_select_step1(
     cfg_t *, cfg_node_t *, cfg_node_t ***, long *, const char *);
-static cfg_rc_t 
+static cfg_rc_t
 cfg_node_select_step2(
     cfg_t *, cfg_node_t *, cfg_node_t ***, long *, const char *, const char *, size_t, long, long, long *);
 
-static cfg_rc_t 
+/* INTERNAL: selection stepping function (part 3, matching node processing) */
+static cfg_rc_t
 cfg_node_select_step3(
-    cfg_t *cfg, 
-    cfg_node_t *node, 
-    cfg_node_t ***result_vec, long *result_len, 
+    cfg_t *cfg,
+    cfg_node_t *node,
+    cfg_node_t ***result_vec, long *result_len,
     const char *spec,
     const char *cpSel, size_t nSel,
     long nFilMin, long nFilMax,
@@ -335,35 +427,47 @@ cfg_node_select_step3(
 {
     cfg_rc_t rc;
 
-    fprintf(stderr, "step3: (1) cpSel=\"%s\", nSel=%d, nFilMin=%ld, nFilMax=%ld, nFound=%ld\n", cpSel, nSel, nFilMin, nFilMax, *nFound);
+#if 0
+    fprintf(stderr, "      step3: (1) cpSel=\"%s\", nSel=%d, spec=\"%s\", nFilMin=%ld, nFilMax=%ld, nFound=%ld\n",
+            cpSel, nSel, spec, nFilMin, nFilMax, *nFound);
+#endif
 
     if (spec[0] == '\0') {
         /* end of selection, node found */
-        *nFound++;
-        fprintf(stderr, "step3: found node=0x%lx!!\n", (unsigned long)node);
+        (*nFound)++;
+#if 0
+        fprintf(stderr, "      step3: (2) found node!!\n");
+        fprintf(stderr, "      step3:     current node=0x%lx type=%s token=\"%s\"\n",
+                (unsigned long)node,
+                (node != NULL ? (node->type == CFG_NODE_TYPE_SEQ ? "SEQ" : (node->type == CFG_NODE_TYPE_DIR ? "DIR" : "ARG")) : "<NULL>/.."),
+                (node != NULL ? (node->token != NULL ? node->token : "<NULL>") : "<NULL>/.."));
+#endif
         if (nFilMin <= *nFound && *nFound <= nFilMax) {
-            if (result_len != NULL)
+            if (result_len != NULL) {
                 (*result_len)++;
-            if (result_vec != NULL) {
-                if ((*result_vec = (cfg_node_t **)realloc(*result_vec, sizeof(cfg_node_t *)*((*result_len)+1))) == NULL)
-                    return CFG_ERR_SYS;
-                (*result_vec)[(*result_len)-1] = node;
-                (*result_vec)[(*result_len)] = NULL;
+                if (result_vec != NULL) {
+                    if ((*result_vec = (cfg_node_t **)realloc(*result_vec, sizeof(cfg_node_t *)*((*result_len)+1))) == NULL)
+                        return CFG_ERR_SYS;
+                    (*result_vec)[(*result_len)-1] = node;
+                    (*result_vec)[(*result_len)] = NULL;
+                }
             }
         }
         rc = CFG_OK;
     }
     else {
+        /* start over with next step */
         rc = cfg_node_select_step1(cfg, node, result_vec, result_len, spec);
     }
     return rc;
 }
 
-static cfg_rc_t 
+/* INTERNAL: selection stepping function (part 2, node matching) */
+static cfg_rc_t
 cfg_node_select_step2(
-    cfg_t *cfg, 
-    cfg_node_t *node, 
-    cfg_node_t ***result_vec, long *result_len, 
+    cfg_t *cfg,
+    cfg_node_t *node,
+    cfg_node_t ***result_vec, long *result_len,
     const char *spec,
     const char *cpSel, size_t nSel,
     long nFilMin, long nFilMax,
@@ -371,99 +475,135 @@ cfg_node_select_step2(
 {
     cfg_rc_t rc;
     char *token;
+    cfg_node_t *node2;
+    cfg_node_type_t type;
 
-    fprintf(stderr, "step2: (1) cpSel=\"%s\", nSel=%d, nFilMin=%ld, nFilMax=%ld, nFound=%ld\n", cpSel, nSel, nFilMin, nFilMax, *nFound);
+#if 0
+    fprintf(stderr, "    step2: (1) cpSel=\"%.*s\", nSel=%d, spec=\"%s\", nFilMin=%ld, nFilMax=%ld, nFound=%ld\n",
+            nSel, cpSel, nSel, spec, nFilMin, nFilMax, *nFound);
+    fprintf(stderr, "    step2:     current node=0x%lx type=%s token=\"%s\"\n",
+            (unsigned long)node,
+            (node != NULL ? (node->type == CFG_NODE_TYPE_SEQ ? "SEQ" : (node->type == CFG_NODE_TYPE_DIR ? "DIR" : "ARG")) : "<NULL>/.."),
+            (node != NULL ? (node->token != NULL ? node->token : "<NULL>") : "<NULL>/.."));
+#endif
 
-    if (strncmp(cpSel, ".", nSel) == 0) {
+    rc = CFG_OK;
+    if (nSel == 1 && strncmp(cpSel, ".", nSel) == 0) {
         /* current node (no-op) */
-        return cfg_node_select_step3(cfg, node, result_vec, result_len, spec, cpSel, nSel, nFilMin, nFilMax, nFound);
+#if 0
+        fprintf(stderr, "    step2: search current node, starting at node 0x%lx\n", (unsigned long)node);
+#endif
+        rc = cfg_node_select_step3(cfg, node, result_vec, result_len, spec, cpSel, nSel, nFilMin, nFilMax, nFound);
     }
-    else if (strncmp(cpSel, "..", nSel) == 0) {
+    else if (nSel == 2 && strncmp(cpSel, "..", nSel) == 0) {
         /* parent node */
-        if ((node = node->parent) == NULL)
-            return CFG_OK;
-        if (node->type == CFG_NODE_TYPE_DIR)
-            if ((node = node->parent) == NULL)
-                return CFG_OK;
-        return cfg_node_select_step3(cfg, node, result_vec, result_len, spec, cpSel, nSel, nFilMin, nFilMax, nFound);
+#if 0
+        fprintf(stderr, "    step2: search parent node, starting at node 0x%lx\n", (unsigned long)node);
+#endif
+        if (cfg_node_get(cfg, node, CFG_NODE_ATTR_PARENT, &node) == CFG_OK && node != NULL)
+            rc = cfg_node_select_step3(cfg, node, result_vec, result_len, spec, cpSel, nSel, nFilMin, nFilMax, nFound);
     }
-    else if (strncmp(cpSel, "....", nSel) == 0) {
+    else if (nSel == 4 && strncmp(cpSel, "....", nSel) == 0) {
         /* anchestor nodes */
-        while ((node = node->parent) != NULL) {
-            if (node->type == CFG_NODE_TYPE_DIR)
-                if ((node = node->parent) == NULL)
-                    break;
+#if 0
+        fprintf(stderr, "    step2: search ancestor nodes, starting at node 0x%lx\n", (unsigned long)node);
+#endif
+        while (cfg_node_get(cfg, node, CFG_NODE_ATTR_PARENT, &node) == CFG_OK && node != NULL)
             if ((rc = cfg_node_select_step3(cfg, node, result_vec, result_len, spec, cpSel, nSel, nFilMin, nFilMax, nFound)) != CFG_OK)
-                return rc;
-        }
-        return CFG_OK;
+                break;
     }
-    else if (strncmp(cpSel, "-", nSel) == 0) {
+    else if (nSel == 1 && strncmp(cpSel, "-", nSel) == 0) {
         /* previous sibling node */
-        if ((rc = cfg_node_get(cfg, node, CFG_NODE_ATTR_LBROTH, &node)) != CFG_OK)
-            return CFG_OK;
-        return cfg_node_select_step3(cfg, node, result_vec, result_len, spec, cpSel, nSel, nFilMin, nFilMax, nFound);
+#if 0
+        fprintf(stderr, "    step2: search previous sibling node, starting at node 0x%lx\n", (unsigned long)node);
+#endif
+        if ((rc = cfg_node_get(cfg, node, CFG_NODE_ATTR_LBROTH, &node)) == CFG_OK && node != NULL)
+            rc = cfg_node_select_step3(cfg, node, result_vec, result_len, spec, cpSel, nSel, nFilMin, nFilMax, nFound);
     }
-    else if (strncmp(cpSel, "--", nSel) == 0) {
+    else if (nSel == 2 && strncmp(cpSel, "--", nSel) == 0) {
         /* preceeding sibling nodes */
-        while ((rc = cfg_node_get(cfg, node, CFG_NODE_ATTR_LBROTH, &node)) == CFG_OK)
+#if 0
+        fprintf(stderr, "    step2: search preceeding sibling nodes, starting at node 0x%lx\n", (unsigned long)node);
+#endif
+        while (cfg_node_get(cfg, node, CFG_NODE_ATTR_LBROTH, &node) == CFG_OK && node != NULL)
             if ((rc = cfg_node_select_step3(cfg, node, result_vec, result_len, spec, cpSel, nSel, nFilMin, nFilMax, nFound)) != CFG_OK)
-                return rc;
-        return CFG_OK;
+                break;
     }
-    else if (strncmp(cpSel, "+", nSel) == 0) {
+    else if (nSel == 1 && strncmp(cpSel, "+", nSel) == 0) {
         /* next sibling node */
-        if ((rc = cfg_node_get(cfg, node, CFG_NODE_ATTR_RBROTH, &node)) != CFG_OK)
-            return CFG_OK;
-        return cfg_node_select_step3(cfg, node, result_vec, result_len, spec, cpSel, nSel, nFilMin, nFilMax, nFound);
+#if 0
+        fprintf(stderr, "    step2: search next sibling node, starting at node 0x%lx\n", (unsigned long)node);
+#endif
+        if ((rc = cfg_node_get(cfg, node, CFG_NODE_ATTR_RBROTH, &node)) == CFG_OK && node != NULL)
+            rc = cfg_node_select_step3(cfg, node, result_vec, result_len, spec, cpSel, nSel, nFilMin, nFilMax, nFound);
     }
-    else if (strncmp(cpSel, "++", nSel) == 0) {
+    else if (nSel == 2 && strncmp(cpSel, "++", nSel) == 0) {
         /* following sibling nodes */
-        while ((rc = cfg_node_get(cfg, node, CFG_NODE_ATTR_RBROTH, &node)) == CFG_OK)
+#if 0
+        fprintf(stderr, "    step2: search following sibling nodes, starting at node 0x%lx\n", (unsigned long)node);
+#endif
+        while (cfg_node_get(cfg, node, CFG_NODE_ATTR_RBROTH, &node) == CFG_OK && node != NULL)
             if ((rc = cfg_node_select_step3(cfg, node, result_vec, result_len, spec, cpSel, nSel, nFilMin, nFilMax, nFound)) != CFG_OK)
-                return rc;
-        return CFG_OK;
+                break;
     }
-    else if (nSel == 0) {
+    else if (nSel == 0 /* "//" */) {
         /* descendant nodes */
-        if ((rc = cfg_node_get(cfg, node, CFG_NODE_ATTR_CHILD1, &node)) == CFG_OK)
-            while ((rc = cfg_node_get(cfg, node, CFG_NODE_ATTR_RBROTH, &node)) == CFG_OK)
+#if 0
+        fprintf(stderr, "    step2: search descendant nodes, starting at node 0x%lx\n", (unsigned long)node);
+#endif
+#if 0 /* FIXME */
+        if ((rc = cfg_node_apply(cfg, node, NULL, NULL, cfg_node_select_step2_descendant, &descendant_ctx)
+            cfg_rc_t cfg_node_select_step2_descendant(cfg_t *cfg, cfg_node_t *node, void *_ctx));
+#endif
+        if ((rc = cfg_node_get(cfg, node, CFG_NODE_ATTR_CHILD1, &node)) == CFG_OK && node != NULL) {
+            do {
+                if ((rc = cfg_node_select_step3(cfg, node, result_vec, result_len, spec, cpSel, nSel, nFilMin, nFilMax, nFound)) != CFG_OK)
+                    return rc;
                 if ((rc = cfg_node_select_step2(cfg, node, result_vec, result_len, spec, cpSel, nSel, nFilMin, nFilMax, nFound)) != CFG_OK)
                     return rc;
+            } while ((rc = cfg_node_get(cfg, node, CFG_NODE_ATTR_RBROTH, &node)) == CFG_OK && node != NULL);
+        }
         return CFG_OK;
     }
     else {
         /* child node */
-        fprintf(stderr, "step2: child node 0x%lx\n", (unsigned long)node);
-        if ((node = node->child1) == NULL)
-            return CFG_OK;
-        if (node->type == CFG_NODE_TYPE_DIR)
-            if ((node = node->child1) == NULL)
-                return CFG_OK;
-        fprintf(stderr, "step2: child node 0x%lx\n", (unsigned long)node);
-        do {
-            if (node->token != NULL) {
-                size_t l;
-                token = node->token;
-                l = strlen(token);
-                fprintf(stderr, "step2: child node: \"%s\"\n", token);
-                if (   (l == 1 && l == nSel && token[0] == '*')
-                    || (l == nSel && strncmp(token, cpSel, nSel) == 0)) {
-                    if ((rc = cfg_node_select_step3(cfg, node, result_vec, result_len, spec, cpSel, nSel, nFilMin, nFilMax, nFound)) != CFG_OK)
-                        return rc;
+#if 0
+        fprintf(stderr, "    step2: search child nodes, starting at node 0x%lx\n", (unsigned long)node);
+#endif
+        if ((rc = cfg_node_get(cfg, node, CFG_NODE_ATTR_CHILD1, &node)) == CFG_OK && node != NULL) {
+            do {
+                if ((rc = cfg_node_get(cfg, node, CFG_NODE_ATTR_TOKEN, &token)) != CFG_OK)
+                    continue;
+                if ((rc = cfg_node_get(cfg, node, CFG_NODE_ATTR_TYPE, &type)) == CFG_OK && type == CFG_NODE_TYPE_DIR) {
+                    if ((rc = cfg_node_get(cfg, node, CFG_NODE_ATTR_CHILD1, &node2)) != CFG_OK || node2 == NULL)
+                        continue;
+                    if ((rc = cfg_node_get(cfg, node2, CFG_NODE_ATTR_TOKEN, &token)) != CFG_OK)
+                        continue;
                 }
-            }
-        } while ((node = node->rbroth) != NULL);
+                if (token != NULL) {
+                    size_t l = strlen(token);
+#if 0
+                    fprintf(stderr, "    step2: compare child node: token \"%s\"\n", token);
+#endif
+                    if (   (l == 1 && nSel == 1 && token[0] == '*')
+                        || (l == nSel && strncmp(token, cpSel, nSel) == 0)) {
+                        if ((rc = cfg_node_select_step3(cfg, node, result_vec, result_len, spec, cpSel, nSel, nFilMin, nFilMax, nFound)) != CFG_OK)
+                            return rc;
+                    }
+                }
+            } while ((rc = cfg_node_get(cfg, node, CFG_NODE_ATTR_RBROTH, &node)) == CFG_OK && node != NULL);
+        }
         return CFG_OK;
     }
     return rc;
 }
 
-static cfg_rc_t 
+/* INTERNAL: selection stepping function (part 1, selection parsing) */
+static cfg_rc_t
 cfg_node_select_step1(
-    cfg_t *cfg, 
-    cfg_node_t *node, 
-    cfg_node_t ***result_vec, long *result_len, 
+    cfg_t *cfg,
+    cfg_node_t *node,
+    cfg_node_t ***result_vec, long *result_len,
     const char *spec)
 {
     const char *cpSel; size_t nSel;
@@ -474,14 +614,16 @@ cfg_node_select_step1(
     cfg_rc_t rc;
     long nFound;
 
-    fprintf(stderr, "step1(spec=\"%s\")\n", spec);
+#if 0
+    fprintf(stderr, "  step1: (0) spec=\"%s\"\n", spec);
+#endif
 
     /* stop processing if spec is empty */
     if (spec[0] == '\0')
         return CFG_OK;
 
     /* determine selection step information */
-    cpSel = spec; 
+    cpSel = spec;
     nSel = strcspn(cpSel, "[/");
     cp = cpSel+nSel;
     nFilMin = 1;
@@ -511,47 +653,63 @@ cfg_node_select_step1(
         cp++;
     spec = cp;
 
-    fprintf(stderr, "    step1: (1) cpSel=\"%s\", nSel=%d, nFilMin=%ld, nFilMax=%ld\n", cpSel, nSel, nFilMin, nFilMax);
+#if 0
+    fprintf(stderr, "  step1: (1) cpSel=\"%s\", nSel=%d, spec=\"%s\", nFilMin=%ld, nFilMax=%ld\n",
+            cpSel, nSel, spec, nFilMin, nFilMax);
+#endif
 
     /* perform pre-selection if filter range is relative to last element */
     if (nFilMin < 0 || nFilMax < 0) {
-        if (nFilMin == -1)
-            nFilMin = LONG_MAX;
-        if (nFilMax == -1)
-            nFilMax = LONG_MAX;
-        if (nFilMin < 0 || nFilMax < 0) {
-            nFound = 0;
-            if ((rc = cfg_node_select_step2(cfg, node, NULL, NULL, spec, cpSel, nSel, 1, LONG_MAX, &nFound)) != CFG_OK)
-                return rc;
-            if (nFilMin < 0) {
-                nFilMin = nFound + nFilMin;
-                if (nFilMin < 1)
-                    nFilMin = 1;
-            }
-            if (nFilMax < 0) {
-                nFilMax = nFound + nFilMax;
-                if (nFilMax < 1)
-                    nFilMax = 1;
-            }
+        nFound = 0;
+#if 0
+        fprintf(stderr, "  step1: pre-selection start\n");
+#endif
+        if ((rc = cfg_node_select_step2(cfg, node, NULL, NULL, spec, cpSel, nSel, 1, LONG_MAX, &nFound)) != CFG_OK)
+            return rc;
+#if 0
+        fprintf(stderr, "  step1: (1b) cpSel=\"%s\", nSel=%d, spec=\"%s\", nFilMin=%ld, nFilMax=%ld nFound=%ld\n", 
+                cpSel, nSel, spec, nFilMin, nFilMax, nFound);
+        fprintf(stderr, "  step1: pre-selection end\n");
+#endif
+        if (nFound == 0)
+            return CFG_OK;
+        if (nFilMin < 0) {
+            nFilMin = nFound + (nFilMin+1);
+            if (nFilMin < 1)
+                nFilMin = 1;
+        }
+        if (nFilMax < 0) {
+            nFilMax = nFound + (nFilMax+1);
+            if (nFilMax < 1)
+                nFilMax = 1;
         }
     }
 
-    fprintf(stderr, "    step1: (2) cpSel=\"%s\", nSel=%d, nFilMin=%ld, nFilMax=%ld\n", cpSel, nSel, nFilMin, nFilMax);
+#if 0
+    fprintf(stderr, "  step1: (2) cpSel=\"%s\", nSel=%d, spec=\"%s\", nFilMin=%ld, nFilMax=%ld\n",
+            cpSel, nSel, spec, nFilMin, nFilMax);
+#endif
 
     /* perform real selection */
     nFound = 0;
     if ((rc = cfg_node_select_step2(cfg, node, result_vec, result_len, spec, cpSel, nSel, nFilMin, nFilMax, &nFound)) != CFG_OK)
         return rc;
 
+#if 0
+    fprintf(stderr, "  step1: (3) cpSel=\"%s\", nSel=%d, spec=\"%s\", nFilMin=%ld, nFilMax=%ld nFound=%ld\n",
+            cpSel, nSel, spec, nFilMin, nFilMax, nFound);
+#endif
+
     return CFG_OK;
 }
 
+/* select a node */
 cfg_rc_t
 cfg_node_select(
-    cfg_t *cfg, 
-    cfg_node_t *node, 
-    cfg_node_t ***result_vec, 
-    const char *fmt, 
+    cfg_t *cfg,
+    cfg_node_t *node,
+    cfg_node_t ***result_vec,
+    const char *fmt,
     ...)
 {
     cfg_rc_t rc;
@@ -560,6 +718,7 @@ cfg_node_select(
     char *cp;
     long result_len;
 
+    /* argument sanity checking */
     if (cfg == NULL || result_vec == NULL || fmt == NULL)
         return CFG_ERR_ARG;
 
@@ -569,15 +728,26 @@ cfg_node_select(
     va_end(ap);
     if (spec == NULL)
         return CFG_ERR_FMT;
+#if 0
+    fprintf(stderr, "select: (1) spec=\"%s\"\n", spec);
+#endif
 
-    /* special case for absolute (start from root-node) selection */
+    /* special cases for absolute (start from root-node) selection */
     cp = spec;
     if (cp[0] == '/') {
+        /* <any,"/x"> is same as <root,"x"> */
+        if (node != NULL)
+            return CFG_ERR_USE;
         node = cfg->root;
         cp++;
     }
-    if (node == NULL)
+    else if (node == NULL) {
+        /* <NULL, "..."> is same as <root, "..."> */
         node = cfg->root;
+    }
+#if 0
+    fprintf(stderr, "select: (2) spec=\"%s\"\n", cp);
+#endif
 
     /* initialize result node array */
     result_len = 0;
@@ -590,60 +760,87 @@ cfg_node_select(
         free(*result_vec);
         return rc;
     }
+#if 0
+    fprintf(stderr, "select: (3) result_len=%ld\n", result_len);
+#endif
 
     return CFG_OK;
 }
 
-cfg_rc_t 
+/* return next matching node (iteration) */
+cfg_rc_t
 cfg_node_find(
-    cfg_t *cfg, 
-    cfg_node_t *node, 
+    cfg_t *cfg,
+    cfg_node_t *node,
     cfg_rc_t (*cb_fct_cmp)(cfg_t *, cfg_node_t *, void *),
-    void *cb_ctx_cmp, 
+    void *cb_ctx_cmp,
     cfg_node_t **cont)
 {
     /* FIXME */
     return CFG_ERR_INT;
 }
 
-cfg_rc_t 
+/* apply for each matching node (recursion) */
+cfg_rc_t
 cfg_node_apply(
-    cfg_t *cfg, 
-    cfg_node_t *node, 
+    cfg_t *cfg,
+    cfg_node_t *node,
     cfg_rc_t (*cb_fct_cmp)(cfg_t *, cfg_node_t *, void *),
     void *cb_ctx_cmp,
     cfg_rc_t (*cb_fct_cb)(cfg_t *, cfg_node_t *, void *),
     void *cb_ctx_cb)
 {
     cfg_rc_t rc;
+    cfg_node_t *child1;
+    cfg_node_t *rbroth;
 
+    /* argument sanity checking */
     if (cfg == NULL)
         return CFG_ERR_ARG;
-    if (node != NULL) {
-        if ((rc = cb_fct_cmp(cfg, node, cb_ctx_cmp)) == CFG_OK)
-            cb_fct_cb(cfg, node, cb_ctx_cb);
-        if (rc != CFG_ERR_NDE)
+
+    /* short-circuit processing */
+    if (node == NULL)
+        return CFG_OK;
+
+    /* remember traversing informations from child and brother
+       (important in case the callbacks destroy it) */
+    child1 = node->child1;
+    rbroth = node->rbroth;
+
+    /* check whether the node matches */
+    if (cb_fct_cmp != NULL)
+        rc = cb_fct_cmp(cfg, node, cb_ctx_cmp);
+    else
+        rc = CFG_OK;
+
+    /* if node matched, apply optional callback on node */
+    if (rc == CFG_OK && cb_fct_cb != NULL)
+        if ((rc = cb_fct_cb(cfg, node, cb_ctx_cb)) != CFG_OK)
             return rc;
-        if (node->child1 != NULL)
-            if ((rc = cfg_node_apply(cfg, node->child1, 
-                                     cb_fct_cmp, cb_ctx_cmp,
-                                     cb_fct_cb, cb_ctx_cb)) != CFG_OK)
-                return rc;
-        if (node->rbroth != NULL)
-            if ((rc = cfg_node_apply(cfg, node->rbroth, 
-                                     cb_fct_cmp, cb_ctx_cmp,
-                                     cb_fct_cb, cb_ctx_cb)) != CFG_OK)
-                return rc;
-    }
+
+    /* recursively apply on child and brother nodes */
+    if (child1 != NULL)
+        if ((rc = cfg_node_apply(cfg, child1,
+                                 cb_fct_cmp, cb_ctx_cmp,
+                                 cb_fct_cb,  cb_ctx_cb)) != CFG_OK)
+            return rc;
+    if (rbroth != NULL)
+        if ((rc = cfg_node_apply(cfg, rbroth,
+                                 cb_fct_cmp, cb_ctx_cmp,
+                                 cb_fct_cb,  cb_ctx_cb)) != CFG_OK)
+            return rc;
+
     return CFG_OK;
 }
 
+/* compare two nodes */
 cfg_rc_t
 cfg_node_cmp(
-    cfg_t *cfg, 
-    cfg_node_t *node, 
+    cfg_t *cfg,
+    cfg_node_t *node,
     void *token)
 {
+    /* argument sanity checking */
     if (cfg == NULL || node == NULL || token == NULL)
         return CFG_ERR_NDE;
     if (node->token == NULL && token == NULL)
@@ -655,10 +852,17 @@ cfg_node_cmp(
     return CFG_ERR_NDE;
 }
 
-cfg_rc_t cfg_node_link(cfg_t *cfg, cfg_node_t *node, cfg_node_attr_t attr, cfg_node_t *node2)
+/* link to nodes together */
+cfg_rc_t
+cfg_node_link(
+    cfg_t *cfg,
+    cfg_node_t *node,
+    cfg_node_attr_t attr,
+    cfg_node_t *node2)
 {
     cfg_node_t *n;
 
+    /* argument sanity checking */
     if (node == NULL || node2 == NULL)
         return CFG_ERR_ARG;
     if (attr == CFG_NODE_ATTR_RBROTH) {
@@ -669,7 +873,7 @@ cfg_rc_t cfg_node_link(cfg_t *cfg, cfg_node_t *node, cfg_node_attr_t attr, cfg_n
             n->parent = node->parent;
             n = n->rbroth;
         }
-        n->rbroth = node->rbroth; 
+        n->rbroth = node->rbroth;
         node->rbroth = node2;
     }
     else if (attr == CFG_NODE_ATTR_CHILD1) {
@@ -684,17 +888,23 @@ cfg_rc_t cfg_node_link(cfg_t *cfg, cfg_node_t *node, cfg_node_attr_t attr, cfg_n
         node->child1 = node2;
     }
     /* FIXME more linkage possibilities */
-    else 
+    else
         return CFG_ERR_ARG;
     return CFG_OK;
 }
 
-cfg_rc_t cfg_node_unlink(cfg_t *cfg, cfg_node_t *node)
+/* unlink a node from others */
+cfg_rc_t
+cfg_node_unlink(
+    cfg_t *cfg,
+    cfg_node_t *node)
 {
     cfg_node_t *n;
 
+    /* argument sanity checking */
     if (node == NULL)
         return CFG_ERR_ARG;
+
     if (node->parent == NULL)
         return CFG_OK;
     if (node->parent->child1 == node) {
