@@ -1,8 +1,8 @@
 /*
 **  OSSP cfg - Configuration Parsing
-**  Copyright (c) 2002-2003 Ralf S. Engelschall <rse@engelschall.com>
-**  Copyright (c) 2002-2003 The OSSP Project (http://www.ossp.org/)
-**  Copyright (c) 2002-2003 Cable & Wireless Deutschland (http://www.cw.com/de/)
+**  Copyright (c) 2002-2005 Ralf S. Engelschall <rse@engelschall.com>
+**  Copyright (c) 2002-2005 The OSSP Project (http://www.ossp.org/)
+**  Copyright (c) 2002-2005 Cable & Wireless (http://www.cw.com/)
 **
 **  This file is part of OSSP cfg, a configuration parsing
 **  library which can be found at http://www.ossp.org/pkg/lib/cfg/.
@@ -25,19 +25,19 @@
 **  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 **  SUCH DAMAGE.
 **
-**  cfg_syn.c: syntax parsing
+**  cfg_syn.c: syntax parsing and formatting
 */
 
 #include <stdio.h>
 #include <ctype.h>
 
 #include "cfg.h"
-#include "cfg_global.h"
 #include "cfg_grid.h"
 #include "cfg_node.h"
 #include "cfg_fmt.h"
 #include "cfg_syn.h"
 #include "cfg_buf.h"
+#include "cfg_global.h"
 
 /* prototypes for Flex-generated scanner */
 extern int  cfg_syn_lex_init(void *);
@@ -57,16 +57,22 @@ cfg_rc_t cfg_syn_import(
     size_t err_len)
 {
     cfg_syn_ctx_t ctx;
+    cfg_buf_t *buf;
+    cfg_rc_t rc;
     void *yyscan;
 
     /* argument sanity checking */
-    if (node == NULL || in_ptr == NULL || in_len == 0)
+    if (node == NULL || in_ptr == NULL)
         return CFG_ERR_ARG;
 
     /* initialize scanner */
     cfg_syn_lex_init(&yyscan);
     cfg_syn_set_extra(&ctx, yyscan);
 
+    /* initialize temporary buffer context */
+    if ((rc = cfg_buf_create(&buf)) != CFG_OK)
+        return rc;
+
     /* establish our own context which is passed
        through the parser and scanner */
     ctx.inputptr = in_ptr;
@@ -75,6 +81,7 @@ cfg_rc_t cfg_syn_import(
     ctx.cfg      = cfg;
     ctx.node     = NULL;
     ctx.rv       = CFG_OK;
+    ctx.buf      = buf;
     ctx.err_buf  = err_buf;
     ctx.err_len  = err_len;
     ctx.yyscan   = yyscan;
@@ -83,6 +90,9 @@ cfg_rc_t cfg_syn_import(
     if (cfg_syn_parse(&ctx))
         ctx.rv = (ctx.rv == CFG_OK ? CFG_ERR_INT : ctx.rv);
 
+    /* destroy temporary buffer */
+    cfg_buf_destroy(buf);
+
     /* destroy scanner */
     cfg_syn_lex_destroy(yyscan);
 
@@ -92,6 +102,37 @@ cfg_rc_t cfg_syn_import(
     return ctx.rv;
 }
 
+/* helper function for copying out input with escaping of characters */
+static size_t cfg_syn_error_cpyout(char *out_ptr, const char *in_ptr, size_t in_len)
+{
+    size_t out_len;
+    char c;
+
+    out_len = 0;
+    while (in_len-- > 0) {
+        c = *in_ptr++;
+        switch (c) {
+            case '\n': *out_ptr++ = '\\'; *out_ptr++ = 'n'; out_len += 2; break;
+            case '\r': *out_ptr++ = '\\'; *out_ptr++ = 'r'; out_len += 2; break;
+            case '\t': *out_ptr++ = '\\'; *out_ptr++ = 't'; out_len += 2; break;
+            case '\b': *out_ptr++ = '\\'; *out_ptr++ = 'b'; out_len += 2; break;
+            case '\f': *out_ptr++ = '\\'; *out_ptr++ = 'f'; out_len += 2; break;
+            default: {
+                if (!isprint((int)c)) {
+                    *out_ptr++ = '\\';
+                    *out_ptr++ = '?';
+                    out_len += 2;
+                }
+                else {
+                    *out_ptr++ = c;
+                    out_len++;
+                }
+            }
+        }
+    }
+    return out_len;
+}
+
 /* remember a parsing error (used internally) */
 void cfg_syn_error(cfg_syn_ctx_t *ctx, cfg_rc_t rv, YYLTYPE *loc, const char *fmt, ...)
 {
@@ -142,40 +183,39 @@ void cfg_syn_error(cfg_syn_ctx_t *ctx, cfg_rc_t rv, YYLTYPE *loc, const char *fm
     }
 
     /* extract token context with mark token borders */
-    if ((cpBuf = malloc((size_t)((cpE-cpP)+2+1))) == NULL)
+    if ((cpBuf = malloc((size_t)(((cpE-cpP)*2)+2+1))) == NULL)
         return;
     cp = cpBuf;
-    n = cpF-cpP;
-    memcpy(cp, cpP, n); cp += n;
+    cp += cfg_syn_error_cpyout(cp, cpP, cpF-cpP);
     *cp++ = '<';
-    n = cpL-cpF;
-    memcpy(cp, cpF, n); cp += n;
+    cp += cfg_syn_error_cpyout(cp, cpF, cpL-cpF);
     *cp++ = '>';
-    n = cpE-cpL;
-    memcpy(cp, cpL, n); cp += n;
+    cp += cfg_syn_error_cpyout(cp, cpL, cpE-cpL);
     *cp++ = '\0';
 
     /* remember parsing error: part I (context part) */
     cfg_fmt_sprintf(ctx->err_buf, ctx->err_len,
-                    "line %d, column %d: `%s'",
+                    "line %d, column %d: `%s': ",
                     line, column, cpBuf);
     free(cpBuf);
 
     /* remember parsing error: part II (parsing part) */
     n = strlen(ctx->err_buf);
     va_start(ap, fmt);
-    cfg_fmt_vsprintf(ctx->err_buf+n, sizeof(ctx->err_len)-n, fmt, ap);
+    cfg_fmt_vsprintf(ctx->err_buf+n, ctx->err_len-n, fmt, ap);
     va_end(ap);
 
     return;
 }
 
+/* internal cfg_syn_export structure */
 typedef struct {
     cfg_t *cfg;
     cfg_buf_t *buf;
     int indent;
 } export_t;
 
+/* internal cfg_syn_export helper function: sprintf(3) style output formatting */
 static void export_format(export_t *ctx, char *fmt, ...)
 {
     int i;
@@ -201,6 +241,7 @@ static void export_format(export_t *ctx, char *fmt, ...)
     return;
 }
 
+/* internal cfg_syn_export helper function: indentation handling */
 static void export_indent(export_t *ctx, signed int n)
 {
     if (n > 0) {
@@ -219,6 +260,7 @@ static void export_indent(export_t *ctx, signed int n)
     }
 }
 
+/* internal cfg_syn_export helper function: token output formatting */
 static void export_token(export_t *ctx, const char *token)
 {
     const char *cp;
@@ -264,6 +306,7 @@ static void export_token(export_t *ctx, const char *token)
     return;
 }
 
+/* internal cfg_syn_export helper function: recursive node processing */
 static void export_node(export_t *ctx, cfg_node_t *node)
 {
     cfg_node_type_t type;
@@ -274,20 +317,22 @@ static void export_node(export_t *ctx, cfg_node_t *node)
     if ((rc = cfg_node_get(ctx->cfg, node, CFG_NODE_ATTR_TYPE, &type)) != CFG_OK)
         return;
     if (type == CFG_NODE_TYPE_SEQ) {
+        /* node is a sequence */
         export_format(ctx, "{\n");
         export_indent(ctx, 1);
         cfg_node_get(ctx->cfg, node, CFG_NODE_ATTR_CHILD1, &node2);
         while (node2 != NULL) {
-            export_node(ctx, node2);
+            export_node(ctx, node2); /* recursion */
             cfg_node_get(ctx->cfg, node2, CFG_NODE_ATTR_RBROTH, &node2);
         }
         export_indent(ctx, -1);
         export_format(ctx, "}");
     }
     else if (type == CFG_NODE_TYPE_DIR) {
+        /* node is a directive */
         cfg_node_get(ctx->cfg, node, CFG_NODE_ATTR_CHILD1, &node2);
         while (node2 != NULL) {
-            export_node(ctx, node2);
+            export_node(ctx, node2); /* recursion */
             cfg_node_get(ctx->cfg, node2, CFG_NODE_ATTR_RBROTH, &node2);
             if (node2 != NULL)
                 export_format(ctx, " ");
@@ -298,6 +343,7 @@ static void export_node(export_t *ctx, cfg_node_t *node)
         export_format(ctx, "\n");
     }
     else if (type == CFG_NODE_TYPE_ARG) {
+        /* node is a token */
         cfg_node_get(ctx->cfg, node, CFG_NODE_ATTR_TOKEN, &token);
         if (token != NULL)
             export_token(ctx, token);
@@ -307,6 +353,7 @@ static void export_node(export_t *ctx, cfg_node_t *node)
     return;
 }
 
+/* export configuration node tree into textual OSSP cfg syntax */
 cfg_rc_t cfg_syn_export(
     cfg_t *cfg,
     cfg_node_t *node,
@@ -326,7 +373,7 @@ cfg_rc_t cfg_syn_export(
     ctx.buf    = buf;
     ctx.indent = 0;
 
-    if ((rc = cfg_node_root(cfg, &root)) != CFG_OK)
+    if ((rc = cfg_node_root(cfg, NULL, &root)) != CFG_OK)
         return rc;
     if (node == root) {
         /* if we dump the whole configuration, treat the first SEQ node special
@@ -348,18 +395,20 @@ cfg_rc_t cfg_syn_export(
     return CFG_OK;
 }
 
+/* internal cfg_syn_destroy helper function: recursive node destruction */
 static cfg_rc_t cfg_syn_destroy_node(cfg_t *cfg, cfg_node_t *node)
 {
     if (node == NULL)
         return CFG_ERR_ARG;
     if (node->child1 != NULL)
-        cfg_syn_destroy_node(cfg, node->child1);
+        cfg_syn_destroy_node(cfg, node->child1); /* recursion */
     if (node->rbroth != NULL)
-        cfg_syn_destroy_node(cfg, node->rbroth);
+        cfg_syn_destroy_node(cfg, node->rbroth); /* recursion */
     cfg_node_destroy(cfg, node);
     return CFG_OK;
 }
 
+/* destroy syntax node tree */
 cfg_rc_t cfg_syn_destroy(cfg_t *cfg, cfg_node_t *node)
 {
     if (node == NULL)
