@@ -1,8 +1,8 @@
 /*
 **  OSSP cfg - Configuration Parsing
-**  Copyright (c) 2002-2003 Ralf S. Engelschall <rse@engelschall.com>
-**  Copyright (c) 2002-2003 The OSSP Project (http://www.ossp.org/)
-**  Copyright (c) 2003-2003 Cable & Wireless Deutschland (http://www.cw.com/de/)
+**  Copyright (c) 2002-2005 Ralf S. Engelschall <rse@engelschall.com>
+**  Copyright (c) 2002-2005 The OSSP Project (http://www.ossp.org/)
+**  Copyright (c) 2003-2005 Cable & Wireless (http://www.cw.com/)
 **
 **  This file is part of OSSP cfg, a configuration parsing
 **  library which can be found at http://www.ossp.org/pkg/lib/cfg/.
@@ -33,7 +33,7 @@
 
 #include "cfg.h"
 
-cfg_rc_t cfg_util_readfile(const char *filename, char **buf_ptr, size_t *buf_len);
+cfg_rc_t cfg_util_readfile(const char *filename, char **buf_ptr, size_t *buf_size, size_t *buf_used);
 
 #endif /* __CFG_UTIL_H__ */
 
