@@ -1,9 +1,10 @@
 /*
-**  OSSP l2 - Logging Library
-**  Copyright (c) 2001-2002 The OSSP Project (http://www.ossp.org/)
-**  Copyright (c) 2001-2002 Cable & Wireless Deutschland (http://www.cw.com/de/)
+**  OSSP l2 - Flexible Logging
+**  Copyright (c) 2001-2005 Cable & Wireless <http://www.cw.com/>
+**  Copyright (c) 2001-2005 The OSSP Project <http://www.ossp.org/>
+**  Copyright (c) 2001-2005 Ralf S. Engelschall <rse@engelschall.com>
 **
-**  This file is part of OSSP L2, a flexible logging library which
+**  This file is part of OSSP l2, a flexible logging library which
 **  can be found at http://www.ossp.org/pkg/lib/l2/.
 **
 **  Permission to use, copy, modify, and distribute this software for
@@ -64,7 +65,7 @@ static l2_result_t hook_create(l2_context_t *ctx, l2_channel_t *ch)
 static l2_result_t hook_configure(l2_context_t *ctx, l2_channel_t *ch, const char *fmt, va_list ap)
 {
     l2_ch_prefix_t *cfg = (l2_ch_prefix_t *)ctx->vp;
-    l2_param_t pa[3]; 
+    l2_param_t pa[3];
     l2_result_t rv;
     l2_env_t *env;
 
@@ -79,7 +80,7 @@ static l2_result_t hook_configure(l2_context_t *ctx, l2_channel_t *ch, const cha
     if (rv == L2_OK) {
         if (cfg->timezone == NULL)
             return L2_ERR_ARG;
-        if (!(   strcmp(cfg->timezone, "local") == 0 
+        if (!(   strcmp(cfg->timezone, "local") == 0
               || strcmp(cfg->timezone, "utc")   == 0))
             return L2_ERR_ARG;
     }
@@ -171,10 +172,12 @@ static l2_result_t hook_write(l2_context_t *ctx, l2_channel_t *ch,
             return L2_ERR_ARG;
         if ((n = strftime(buf2, sizeof(buf2), buf1, tm)) == 0)
             return L2_ERR_SYS;
+        n += l2_util_sprintf(buf2+n, sizeof(buf2)-n, "%s", buf);
         downstream = NULL;
         while ((rv = l2_channel_downstream(ch, &downstream)) == L2_OK)
             if ((rv = l2_channel_write(downstream, level, buf2, n)) != L2_OK)
                 return rv;
+        return L2_OK;
     }
 
     return L2_OK_PASS;
